@@ -185,19 +185,19 @@ class Database(object):
             self.cancel()
             raise
 
-    def retry_if_modified(self, operation, limit=RETRY_LIMIT):
+    def retry_if_modified(self, operation, limit=RETRY_LIMIT, refresh=True):
         tries = 0
         while True:
             try:
                 return operation()
-            except xapian.DatabaseModifiedError:
-                if tries > limit:
+            except xapian.DatabaseError, e:
+                if tries >= limit:
                     logger.warning(
-                        'Database operation failed after %s retries.',
-                        tries)
+                        '%s after %s retries, failing.', e, tries)
                     raise
-                logger.info('Retrying database operation.')
-                self.reopen()
+                logger.info('%s: after %s retries, retrying', e, tries)
+                time.sleep(tries * .1)
+                self.reopen(refresh=refresh)
                 tries += 1
 
     def __init__(self, path,
@@ -210,6 +210,7 @@ class Database(object):
         self._overwrite = overwrite
         self.spelling = spelling
         self.replicated = replicated
+        self.type_map = {}
 
         if isinstance(path, basestring):
             if writable:
@@ -229,12 +230,7 @@ class Database(object):
                                 "if a database path is provided.")
             self.backend = self.db_path
 
-        self.relevance_prefixes = {}
-        self.boolean_prefixes = {}
-        self.values = {}
-        self.value_sorts = {}
-        self.type_map = {}
-        self.refresh_metadata()
+        self.reopen()
 
     def close(self):
         self.backend.close()
@@ -300,13 +296,34 @@ class Database(object):
 
     def __len__(self):
         """ Return the number of documents in this database. """
-        self.backend.reopen()
+        self.reopen()
         return self.backend.get_doccount()
 
-    def allterms(self, prefix=""):
-        self.backend.reopen()
-        for t in self.backend.allterms(prefix):
-            yield t.term
+    def allterms(self, prefix="", retry_limit=RETRY_LIMIT):
+        self.reopen()
+        seen = set()
+        tries = 0
+        # we can't use retry_if_modified because this
+        # is not an atomic operation that returns one result
+        while True:
+            try:
+                for t in self.backend.allterms(prefix):
+                    term = t.term
+                    if term in seen:
+                        continue
+                    seen.add(term)
+                    yield term
+                break
+            except xapian.DatabaseError:
+                if tries > retry_limit:
+                    logger.warning(
+                        'allterms() failed after %s retries.',
+                        tries)
+                    raise
+                logger.info('Retrying allterms() operation.')
+                time.sleep(tries * .1)
+                self.reopen()
+                tries += 1
 
     def get(self, docid, default=None):
         """ Get a document with the given docid, or the default value if
@@ -458,7 +475,7 @@ class Database(object):
             doc.set_data(data)
         return doc
 
-    def reopen(self):
+    def reopen(self, retry_limit=RETRY_LIMIT, refresh=True):
         """
         Reopen the database.  Called before most query methods.  If
         replication is used, the db is closed and reopened.
@@ -472,53 +489,48 @@ class Database(object):
                                   "using replication.")
             self.close()
             self.backend = xapian.Database(self.db_path)
-        self.refresh_metadata()
 
-    def refresh_metadata(self):
-        """Refresh the local connection's view of the prefix and value
-        data.
-        """
-        try:
-            keys = self.backend.metadata_keys()
-        except UnimplementedError:
-            # Inmemory backends don't expose metadata, so just return
-            # since they can't persist anything anyway!
-            return
+        if refresh:
+            try:
+                op = lambda: self.backend.metadata_keys()
+                # don't recurse into refresh here, just reopen and retry
+                keys = self.retry_if_modified(op, retry_limit, False)
+            except UnimplementedError:
+                # Inmemory backends don't expose metadata, so just return
+                # since they can't persist anything anyway!
+                return
 
-        self.relevance_prefixes.clear()
-        self.boolean_prefixes.clear()
-        self.values.clear()
-        self.value_sorts.clear()
+            self.relevance_prefixes = {}
+            self.boolean_prefixes = {}
+            self.values = {}
+            self.value_sorts = {}
 
-        for k in keys:
-            if k.startswith(self.relevance_prefix):
-                prefix = k[len(self.relevance_prefix):]
-                self.relevance_prefixes[prefix] = self.backend.get_metadata(k)
-            elif k.startswith(self.boolean_prefix):
-                prefix = k[len(self.boolean_prefix):]
-                self.boolean_prefixes[prefix] = self.backend.get_metadata(k)
-            elif k.startswith(self.value_prefix):
-                value = k[len(self.value_prefix):]
-                self.values[value] = int(self.backend.get_metadata(k))
-            elif k.startswith(self.value_sort_prefix):
-                value = k[len(self.value_sort_prefix):]
-                self.value_sorts[value] = self.backend.get_metadata(k)
-
-        try:
-            # hit the property to refresh this value
-            count = self.value_count
-            count = count # pyflakes
-        except ValueError:
-            if self._writable:
-                self.value_count = 0
-
-    # FIXME:mp the transactional methods noop for backends that don't
-    # support transaction (only inmemory to date) perhaps support for
-    # those backends should be dropped.
+            for k in keys:
+                op = lambda: self.backend.get_metadata(k)
+                val = self.retry_if_modified(op, retry_limit, False)
+                if k.startswith(self.relevance_prefix):
+                    prefix = k[len(self.relevance_prefix):]
+                    self.relevance_prefixes[prefix] = val
+                elif k.startswith(self.boolean_prefix):
+                    prefix = k[len(self.boolean_prefix):]
+                    self.boolean_prefixes[prefix] = val
+                elif k.startswith(self.value_prefix):
+                    value = k[len(self.value_prefix):]
+                    self.values[value] = int(val)
+                elif k.startswith(self.value_sort_prefix):
+                    value = k[len(self.value_sort_prefix):]
+                    self.value_sorts[value] = val
+            try:
+                # hit the property to refresh this value
+                count = self.value_count
+                count = count # pyflakes
+            except ValueError:
+                if self._writable:
+                    self.value_count = 0
 
     def begin(self):
         if self._writable:
-            self.backend.reopen()
+            self.reopen()
             try:
                 self.backend.begin_transaction()
             except Exception:
@@ -575,7 +587,8 @@ class Database(object):
                 language=None,
                 translit=None,
                 default_op=Query.OP_AND,
-                parser_flags=default_parser_flags):
+                parser_flags=default_parser_flags,
+                retry_limit=RETRY_LIMIT):
         """Return a query object, constructed from a string, query
         object, or mixed sequence of any number of strings, query
         objects, or subsequences of the same form.
@@ -598,8 +611,7 @@ class Database(object):
                     query = query.encode(translit)
                 qp = self.prepare_query_parser(language, default_op)
                 op = lambda: qp.parse_query(query, parser_flags)
-                return self.retry_if_modified(op)
-
+                return self.retry_if_modified(op, retry_limit)
         else:
             return reduce(partial(Query, default_op),
                           (self.querify(q,
@@ -621,7 +633,8 @@ class Database(object):
               document=False,
               echo=False,
               parser_flags=default_parser_flags,
-              default_op=Query.OP_AND):
+              default_op=Query.OP_AND,
+              retry_limit=RETRY_LIMIT):
         """
         Query the database with the provided string or xapian Query
         object.  A string is passed into xapians QueryParser first to
@@ -643,11 +656,51 @@ class Database(object):
         if echo:
             start = time.time()
             print "Fetching mset..."
-        mset = self._build_mset(enq, offset, limit, order, reverse,
-                                check, match_decider, match_spy)
-        if echo:
-            print "Fetched mset in %s" % str(time.time() - start)
-        return self._return_objects(mset, document=document)
+
+        # convoluted logic here is to retry queries that die in the
+        # middle of result iteration because the db was closed (due to
+        # replication).
+
+        tries = 0
+        seen = set()
+        while True:
+            try:
+                # _build_mset may retry internally on DatabaseError
+                mset = self._build_mset(enq, offset, limit, order, reverse,
+                                        check, match_decider, match_spy,
+                                        retry_limit=retry_limit)
+                if echo:
+                    print "Fetched mset in %s" % str(time.time() - start)
+
+                for record in mset:
+                    doc = record.document
+                    docid = doc.get_docid()
+                    if docid in seen:
+                        continue
+                    if document:
+                        seen.add(docid)
+                        yield doc
+                    else:
+                        # retry getting the actual document data
+                        op = lambda: self.backend.get_document(docid).get_data()
+                        data = self.retry_if_modified(op, retry_limit)
+                        typ, data = loads(data)
+                        seen.add(docid)
+                        yield self.record_factory(_lookup_schema(typ).from_flat(data))
+                # no errors exhuasting the set? break out and we're done
+                break
+            except xapian.DatabaseError, e:
+                # an error occured, either, the db was closed, or the
+                # modified error happened two frequently in the inner
+                # loop, so we are going to replay the whole query
+                if tries > retry_limit:
+                    logger.warning(
+                        'Database replay failed after %s retries.',
+                        tries)
+                    raise
+                logger.info('Replaying database query.')
+                self.reopen()
+                tries += 1
 
     def count(self,
               query="",
@@ -655,7 +708,8 @@ class Database(object):
               echo=False,
               translit=None,
               parser_flags=default_parser_flags,
-              default_op=Query.OP_AND):
+              default_op=Query.OP_AND,
+              retry_limit=RETRY_LIMIT):
         """
         Query the database with the provided string or xapian Query
         object.  A string is passed into xapians QueryParser first to
@@ -668,7 +722,7 @@ class Database(object):
         enq = xapian.Enquire(self.backend)
         enq.set_query(query)
 
-        mset = self._build_mset(enq)
+        mset = self._build_mset(enq, retry_limit=retry_limit)
         return mset.size()
 
     def facet(self, query,
@@ -679,7 +733,8 @@ class Database(object):
               mlimit=0,
               klimit=1.0,
               kmlimit=1.0,
-              echo=False):
+              echo=False,
+              retry_limit=RETRY_LIMIT):
         """Get facet suggestions for the query, then the query with
         each suggested facet, asking xapian for an estimated count of
         each sub-query.
@@ -697,9 +752,10 @@ class Database(object):
                                    mlimit=mlimit,
                                    klimit=klimit,
                                    kmlimit=kmlimit,
-                                   echo=echo)
+                                   echo=echo,
+                                   retry_limit=retry_limit)
         for facet in suggestions:
-            q = self.querify([query, facet])
+            q = self.querify([query, facet], retry_limit=retry_limit)
             if echo:
                 print str(q)
             else:
@@ -711,12 +767,14 @@ class Database(object):
                echo=False,
                translit=None,
                default_op=Query.OP_AND,
-               parser_flags=default_parser_flags):
+               parser_flags=default_parser_flags,
+               retry_limit=RETRY_LIMIT):
         """
         Expand a query on a given set of prefixes.
         """
 
-        q = self.querify(query, language, translit, default_op, parser_flags)
+        q = self.querify(query, language, translit, default_op, parser_flags,
+                         retry_limit=retry_limit)
         if echo:
             print q
         results = {}
@@ -730,16 +788,18 @@ class Database(object):
         r = self.facet(
             q, prefix=head, language=language, echo=echo,
             limit=limit, mlimit=mlimit,
-            klimit=klimit, kmlimit=kmlimit).items()
+            klimit=klimit, kmlimit=kmlimit,
+            retry_limit=retry_limit).items()
 
         for name, score in r:
             if tail:
-                subq = self.querify([q, name])
+                subq = self.querify([q, name], retry_limit=retry_limit)
                 if echo:
                     print subq
                 r = self.expand(
                     subq, tail, language, echo,
-                    default_op, parser_flags)
+                    default_op, parser_flags,
+                    retry_limit=retry_limit)
                 results[(name, score)] = r
             else:
                 results[(name, score)] = score
@@ -750,9 +810,9 @@ class Database(object):
                  klimit=1.0,
                  language=None,
                  translit=None,
-                 check_at_least=None,
                  default_op=Query.OP_AND,
-                 parser_flags=default_parser_flags):
+                 parser_flags=default_parser_flags,
+                 retry_limit=RETRY_LIMIT):
         """Estimate the number of documents that will be yielded with the
         given query.
 
@@ -766,13 +826,12 @@ class Database(object):
             limit = int(self.backend.get_doccount() * klimit)
 
         query = self.querify(query, language, translit,
-                             default_op, parser_flags)
+                             default_op, parser_flags,
+                             retry_limit=retry_limit)
 
         enq.set_query(query)
-        if check_at_least is not None:
-            mset = enq.get_mset(0, 0, limit)
-        else:
-            mset = enq.get_mset(0, 0, limit)
+        op = lambda: enq.get_mset(0, 0, limit)
+        mset = self.retry_if_modified(op, retry_limit)
             
         return mset.get_matches_estimated()
 
@@ -781,28 +840,30 @@ class Database(object):
         Return a count of the number of documents indexed for a given
         term.  Useful for testing.
         """
-        self.backend.reopen()
+        self.reopen()
         return self.backend.get_termfreq(term)
 
     def describe_query(self, query,
                        language=None,
-                       default_op=Query.OP_AND):
+                       default_op=Query.OP_AND,
+                       retry_limit=RETRY_LIMIT):
         """
         Describe the parsed query.
         """
         qp = self.prepare_query_parser(language, default_op)
         op = lambda: qp.parse_query(query, default_parser_flags) 
-        return str(self.retry_if_modified(op))
+        return str(self.retry_if_modified(op, retry_limit))
 
     def spell(self, query,
-              language=None):
+              language=None,
+              retry_limit=RETRY_LIMIT):
         """
         Suggest a query string with corrected spelling.
         """
-        self.backend.reopen()
+        self.reopen()
         qp = self.prepare_query_parser(language)
         op = lambda: qp.parse_query(query, QueryParser.FLAG_SPELLING_CORRECTION)
-        self.retry_if_modified(op)
+        self.retry_if_modified(op, retry_limit)
         return qp.get_corrected_query_string().decode('utf8')
 
     def suggest(self, query,
@@ -819,7 +880,8 @@ class Database(object):
                 score=False,
                 echo=False,
                 default_op=Query.OP_AND,
-                parser_flags=default_parser_flags):
+                parser_flags=default_parser_flags,
+                retry_limit=RETRY_LIMIT):
         """
         Suggest terms that would possibly yield more relevant results
         for the given query.
@@ -836,15 +898,12 @@ class Database(object):
             print str(query)
         enq.set_query(query)
 
-        mset = self._build_mset(enq, offset=moffset, limit=mlimit)
+        mset = self._build_mset(enq, offset=moffset, limit=mlimit,
+                                retry_limit=retry_limit)
 
         rset = xapian.RSet()
         for m in mset:
-            if hasattr(m, 'docid'): # 1.1
-                docid = m.docid
-            else:
-                docid = m[xapian.MSET_DID] # 1.0
-            rset.add_document(docid)
+            rset.add_document(m.docid)
 
         if prefix is not None:
             decider = PrefixDecider(prefix)
@@ -878,7 +937,8 @@ class Database(object):
                     reverse=False,
                     check=None,
                     match_decider=None,
-                    match_spy=None):
+                    match_spy=None,
+                    retry_limit=RETRY_LIMIT):
         if order is not None:
             if isinstance(order, basestring):
                 try:
@@ -895,21 +955,7 @@ class Database(object):
 
         op = lambda: enq.get_mset(
                 offset, limit, check, None, match_decider, match_spy)
-        return self.retry_if_modified(op)
-
-    def _return_objects(self, mset, document=False):
-        for record in mset:
-            if hasattr(record, 'document'):  # 1.1
-                doc = record.document
-            else:
-                doc = record[xapian.MSET_DOCUMENT] # 1.0
-            if document:
-                yield doc
-            else:
-                op = lambda: doc.get_data()
-                data = self.retry_if_modified(op)
-                typ, data = loads(data)
-                yield self.record_factory(_lookup_schema(typ).from_flat(data))
+        return self.retry_if_modified(op, retry_limit)
 
 
 def jsonrpc_wrapper(f):
